@@ -1,18 +1,22 @@
-#  Copyright 2022 Upstream Data Inc
-#
-#  Licensed under the Apache License, Version 2.0 (the "License");
-#  you may not use this file except in compliance with the License.
-#  You may obtain a copy of the License at
-#
-#      http://www.apache.org/licenses/LICENSE-2.0
-#
-#  Unless required by applicable law or agreed to in writing, software
-#  distributed under the License is distributed on an "AS IS" BASIS,
-#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
-#  See the License for the specific language governing permissions and
-#  limitations under the License.
+# ------------------------------------------------------------------------------
+#  Copyright 2022 Upstream Data Inc                                            -
+#                                                                              -
+#  Licensed under the Apache License, Version 2.0 (the "License");             -
+#  you may not use this file except in compliance with the License.            -
+#  You may obtain a copy of the License at                                     -
+#                                                                              -
+#      http://www.apache.org/licenses/LICENSE-2.0                              -
+#                                                                              -
+#  Unless required by applicable law or agreed to in writing, software         -
+#  distributed under the License is distributed on an "AS IS" BASIS,           -
+#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.    -
+#  See the License for the specific language governing permissions and         -
+#  limitations under the License.                                              -
+# ------------------------------------------------------------------------------
+import asyncio
+import logging
 
-from pyasic.API import BaseMinerAPI
+from pyasic.API import APIError, BaseMinerAPI
 
 
 class BMMinerAPI(BaseMinerAPI):
@@ -33,8 +37,41 @@ class BMMinerAPI(BaseMinerAPI):
         port: The port to reference the API on.  Default is 4028.
     """
 
-    def __init__(self, ip: str, port: int = 4028) -> None:
-        super().__init__(ip, port)
+    def __init__(self, ip: str, api_ver: str = "0.0.0", port: int = 4028) -> None:
+        super().__init__(ip, port=port)
+        self.api_ver = api_ver
+
+    async def multicommand(self, *commands: str, allow_warning: bool = True) -> dict:
+        # make sure we can actually run each command, otherwise they will fail
+        commands = self._check_commands(*commands)
+        # standard multicommand format is "command1+command2"
+        # doesn't work for S19 which uses the backup _x19_multicommand
+        command = "+".join(commands)
+        try:
+            data = await self.send_command(command, allow_warning=allow_warning)
+        except APIError:
+            logging.debug(f"{self} - (Multicommand) - Handling X19 multicommand.")
+            data = await self._x19_multicommand(
+                *command.split("+"), allow_warning=allow_warning
+            )
+        data["multicommand"] = True
+        return data
+
+    async def _x19_multicommand(self, *commands, allow_warning: bool = True) -> dict:
+        tasks = []
+        # send all commands individually
+        for cmd in commands:
+            tasks.append(
+                asyncio.create_task(self._handle_multicommand(cmd, allow_warning=True))
+            )
+
+        all_data = await asyncio.gather(*tasks)
+
+        data = {}
+        for item in all_data:
+            data.update(item)
+
+        return data
 
     async def version(self) -> dict:
         """Get miner version info.
