@@ -1,40 +1,43 @@
-#  Copyright 2022 Upstream Data Inc
-#
-#  Licensed under the Apache License, Version 2.0 (the "License");
-#  you may not use this file except in compliance with the License.
-#  You may obtain a copy of the License at
-#
-#      http://www.apache.org/licenses/LICENSE-2.0
-#
-#  Unless required by applicable law or agreed to in writing, software
-#  distributed under the License is distributed on an "AS IS" BASIS,
-#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
-#  See the License for the specific language governing permissions and
-#  limitations under the License.
+# ------------------------------------------------------------------------------
+#  Copyright 2022 Upstream Data Inc                                            -
+#                                                                              -
+#  Licensed under the Apache License, Version 2.0 (the "License");             -
+#  you may not use this file except in compliance with the License.            -
+#  You may obtain a copy of the License at                                     -
+#                                                                              -
+#      http://www.apache.org/licenses/LICENSE-2.0                              -
+#                                                                              -
+#  Unless required by applicable law or agreed to in writing, software         -
+#  distributed under the License is distributed on an "AS IS" BASIS,           -
+#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.    -
+#  See the License for the specific language governing permissions and         -
+#  limitations under the License.                                              -
+# ------------------------------------------------------------------------------
 
 import asyncio
-import re
-import json
-import hashlib
-import binascii
 import base64
+import binascii
+import datetime
+import hashlib
+import json
 import logging
+import re
 from typing import Union
 
-from passlib.handlers.md5_crypt import md5_crypt
 from cryptography.hazmat.primitives.ciphers import Cipher, algorithms, modes
+from passlib.handlers.md5_crypt import md5_crypt
 
-from pyasic.errors import APIError
 from pyasic.API import BaseMinerAPI
+from pyasic.errors import APIError
+from pyasic.misc import api_min_version
 from pyasic.settings import PyasicSettings
 
-
 ### IMPORTANT ###
 # you need to change the password of the miners using the Whatsminer
 # tool, then you can set them back to admin with this tool, but they
 # must be changed to something else and set back to admin with this
 # or the privileged API will not work using admin as the password.  If
-# you change the password, you can pass that to the this class as pwd,
+# you change the password, you can pass that to this class as pwd,
 # or add it as the Whatsminer_pwd in the settings.toml file.
 
 
@@ -83,7 +86,7 @@ def _add_to_16(string: str) -> bytes:
 def parse_btminer_priviledge_data(token_data: dict, data: dict):
     """Parses data returned from the BTMiner privileged API.
 
-    Parses data from the BTMiner privileged API using the the token
+    Parses data from the BTMiner privileged API using the token
     from the API in an AES format.
 
     Parameters:
@@ -125,6 +128,7 @@ def create_privileged_cmd(token_data: dict, command: dict) -> bytes:
     Returns:
         The encrypted privileged command to be sent to the miner.
     """
+    logging.debug(f"(Create Prilileged Command) - Creating Privileged Command")
     # add token to command
     command["token"] = token_data["host_sign"]
     # encode host_passwd data and get hexdigest
@@ -180,64 +184,92 @@ class BTMinerAPI(BaseMinerAPI):
     def __init__(
         self,
         ip: str,
+        api_ver: str = "0.0.0",
         port: int = 4028,
         pwd: str = PyasicSettings().global_whatsminer_password,
     ):
         super().__init__(ip, port)
         self.pwd = pwd
         self.current_token = None
+        self.api_ver = api_ver
 
-    async def send_command(
+    async def multicommand(self, *commands: str, allow_warning: bool = True) -> dict:
+        """Creates and sends multiple commands as one command to the miner.
+
+        Parameters:
+            *commands: The commands to send as a multicommand to the miner.
+            allow_warning: A boolean to supress APIWarnings.
+        """
+        # make sure we can actually run each command, otherwise they will fail
+        commands = self._check_commands(*commands)
+        # standard multicommand format is "command1+command2"
+        # commands starting with "get_" and the "status" command aren't supported, but we can fake that
+
+        tasks = []
+
+        for command in list(commands):
+            if command.startswith("get_") or command == "status":
+                commands.remove(command)
+                # send seperately and append later
+                tasks.append(
+                    asyncio.create_task(
+                        self._handle_multicommand(command, allow_warning=allow_warning)
+                    )
+                )
+
+        command = "+".join(commands)
+        tasks.append(
+            asyncio.create_task(
+                self._handle_multicommand(command, allow_warning=allow_warning)
+            )
+        )
+
+        all_data = await asyncio.gather(*tasks)
+
+        logging.debug(f"{self} - (Multicommand) - Received data")
+
+        data = {}
+        for item in all_data:
+            data.update(item)
+
+        data["multicommand"] = True
+        return data
+
+    async def send_privileged_command(
         self,
         command: Union[str, bytes],
-        parameters: Union[str, int, bool] = None,
         ignore_errors: bool = False,
+        timeout: int = 10,
         **kwargs,
     ) -> dict:
-        # check if command is a string
-        # if its bytes its encoded and needs to be sent raw
-        if isinstance(command, str):
-            # if it is a string, put it into the standard command format
-            command = json.dumps({"command": command}).encode("utf-8")
+        logging.debug(
+            f"{self} - (Send Privileged Command) - {command} " + f"with args {kwargs}"
+            if len(kwargs) > 0
+            else ""
+        )
+        command = {"cmd": command, **kwargs}
+
+        token_data = await self.get_token()
+        enc_command = create_privileged_cmd(token_data, command)
+
+        logging.debug(f"{self} - (Send Privileged Command) - Sending")
         try:
-            # get reader and writer streams
-            reader, writer = await asyncio.open_connection(str(self.ip), self.port)
-        # handle OSError 121
-        except OSError as e:
-            if e.winerror == "121":
-                print("Semaphore Timeout has Expired.")
-            return {}
-
-        # send the command
-        writer.write(command)
-        await writer.drain()
-
-        # instantiate data
-        data = b""
-
-        # loop to receive all the data
-        try:
-            while True:
-                d = await reader.read(4096)
-                if not d:
-                    break
-                data += d
-        except Exception as e:
-            logging.info(f"{str(self.ip)}: {e}")
+            data = await self._send_bytes(enc_command, timeout)
+        except (asyncio.CancelledError, asyncio.TimeoutError):
+            if ignore_errors:
+                return {}
+            raise APIError("No data was returned from the API.")
 
+        if not data:
+            if ignore_errors:
+                return {}
+            raise APIError("No data was returned from the API.")
         data = self._load_api_data(data)
 
-        # close the connection
-        writer.close()
-        await writer.wait_closed()
-
-        # check if the returned data is encoded
-        if "enc" in data.keys():
-            # try to parse the encoded data
-            try:
-                data = parse_btminer_priviledge_data(self.current_token, data)
-            except Exception as e:
-                logging.info(f"{str(self.ip)}: {e}")
+        try:
+            data = parse_btminer_priviledge_data(self.current_token, data)
+        except Exception as e:
+            logging.info(f"{str(self.ip)}: {e}")
 
         if not ignore_errors:
             # if it fails to validate, it is likely an error
@@ -257,6 +289,13 @@ class BTMinerAPI(BaseMinerAPI):
             An encoded token and md5 password, which are used for the privileged API.
         </details>
         """
+        logging.debug(f"{self} - (Get Token) - Getting token")
+        if self.current_token:
+            if self.current_token[
+                "timestamp"
+            ] > datetime.datetime.now() - datetime.timedelta(minutes=30):
+                return self.current_token
+
         # get the token
         data = await self.send_command("get_token")
 
@@ -278,7 +317,11 @@ class BTMinerAPI(BaseMinerAPI):
         self.current_token = {
             "host_sign": host_sign,
             "host_passwd_md5": host_passwd_md5,
+            "timestamp": datetime.datetime.now(),
         }
+        logging.debug(
+            f"{self} - (Get Token) - Gathered token data: {self.current_token}"
+        )
         return self.current_token
 
     #### PRIVILEGED COMMANDS ####
@@ -320,46 +363,18 @@ class BTMinerAPI(BaseMinerAPI):
             A dict from the API to confirm the pools were updated.
         </details>
         """
-        # get the token and password from the miner
-        token_data = await self.get_token()
-
-        # parse pool data
-        if not pool_1:
-            raise APIError("No pools set.")
-        elif pool_2 and pool_3:
-            command = {
-                "cmd": "update_pools",
-                "pool1": pool_1,
-                "worker1": worker_1,
-                "passwd1": passwd_1,
-                "pool2": pool_2,
-                "worker2": worker_2,
-                "passwd2": passwd_2,
-                "pool3": pool_3,
-                "worker3": worker_3,
-                "passwd3": passwd_3,
-            }
-        elif pool_2:
-            command = {
-                "cmd": "update_pools",
-                "pool1": pool_1,
-                "worker1": worker_1,
-                "passwd1": passwd_1,
-                "pool2": pool_2,
-                "worker2": worker_2,
-                "passwd2": passwd_2,
-            }
-        else:
-            command = {
-                "cmd": "update_pools",
-                "pool1": pool_1,
-                "worker1": worker_1,
-                "passwd1": passwd_1,
-            }
-        # encode the command with the token data
-        enc_command = create_privileged_cmd(token_data, command)
-        # send the command
-        return await self.send_command(enc_command)
+        return await self.send_privileged_command(
+            "update_pools",
+            pool1=pool_1,
+            worker1=worker_1,
+            passwd1=passwd_1,
+            pool2=pool_2,
+            worker2=worker_2,
+            passwd2=passwd_2,
+            pool3=pool_3,
+            worker3=worker_3,
+            passwd3=passwd_3,
+        )
 
     async def restart(self) -> dict:
         """Restart BTMiner using the API.
@@ -373,10 +388,7 @@ class BTMinerAPI(BaseMinerAPI):
             A reply informing of the restart.
         </details>
         """
-        command = {"cmd": "restart_btminer"}
-        token_data = await self.get_token()
-        enc_command = create_privileged_cmd(token_data, command)
-        return await self.send_command(enc_command)
+        return await self.send_privileged_command("restart_btminer")
 
     async def power_off(self, respbefore: bool = True) -> dict:
         """Power off the miner using the API.
@@ -393,12 +405,8 @@ class BTMinerAPI(BaseMinerAPI):
         </details>
         """
         if respbefore:
-            command = {"cmd": "power_off", "respbefore": "true"}
-        else:
-            command = {"cmd": "power_off", "respbefore": "false"}
-        token_data = await self.get_token()
-        enc_command = create_privileged_cmd(token_data, command)
-        return await self.send_command(enc_command)
+            return await self.send_privileged_command("power_off", respbefore="true")
+        return await self.send_privileged_command("power_off", respbefore="false")
 
     async def power_on(self) -> dict:
         """Power on the miner using the API.
@@ -409,14 +417,10 @@ class BTMinerAPI(BaseMinerAPI):
         the password of the miner using the Whatsminer tool.
 
         Returns:
-
             A reply informing of the status of powering on.
         </details>
         """
-        command = {"cmd": "power_on"}
-        token_data = await self.get_token()
-        enc_command = create_privileged_cmd(token_data, command)
-        return await self.send_command(enc_command)
+        return await self.send_privileged_command("power_on")
 
     async def reset_led(self) -> dict:
         """Reset the LED on the miner using the API.
@@ -427,14 +431,10 @@ class BTMinerAPI(BaseMinerAPI):
         changing the password of the miner using the Whatsminer tool.
 
         Returns:
-
             A reply informing of the status of resetting the LED.
         </details>
         """
-        command = {"cmd": "set_led", "param": "auto"}
-        token_data = await self.get_token()
-        enc_command = create_privileged_cmd(token_data, command)
-        return await self.send_command(enc_command)
+        return await self.set_led(auto=True)
 
     async def set_led(
         self,
@@ -458,23 +458,14 @@ class BTMinerAPI(BaseMinerAPI):
             duration: LED on time in the cycle in ms.
             start: LED on time offset in the cycle in ms.
         Returns:
-
             A reply informing of the status of setting the LED.
         </details>
         """
-        if not auto:
-            command = {
-                "cmd": "set_led",
-                "color": color,
-                "period": period,
-                "duration": duration,
-                "start": start,
-            }
-        else:
-            command = {"cmd": "set_led", "param": "auto"}
-        token_data = await self.get_token()
-        enc_command = create_privileged_cmd(token_data, command)
-        return await self.send_command(enc_command, ignore_errors=True)
+        if auto:
+            return await self.send_privileged_command("set_led", param="auto")
+        return await self.send_privileged_command(
+            "set_led", color=color, period=period, duration=duration, start=start
+        )
 
     async def set_low_power(self) -> dict:
         """Set low power mode on the miner using the API.
@@ -485,14 +476,10 @@ class BTMinerAPI(BaseMinerAPI):
         changing the password of the miner using the Whatsminer tool.
 
         Returns:
-
             A reply informing of the status of setting low power mode.
         </details>
         """
-        command = {"cmd": "set_low_power"}
-        token_data = await self.get_token()
-        enc_command = create_privileged_cmd(token_data, command)
-        return await self.send_command(enc_command)
+        return await self.send_privileged_command("set_low_power")
 
     async def update_firmware(self):  # noqa - static
         """Not implemented."""
@@ -500,20 +487,23 @@ class BTMinerAPI(BaseMinerAPI):
         # requires a file stream in bytes
         return NotImplementedError
 
-    async def reboot(self) -> dict:
+    async def reboot(self, timeout: int = 10) -> dict:
         """Reboot the miner using the API.
         <details>
             <summary>Expand</summary>
 
         Returns:
-
             A reply informing of the status of the reboot.
         </details>
         """
-        command = {"cmd": "reboot"}
-        token_data = await self.get_token()
-        enc_command = create_privileged_cmd(token_data, command)
-        return await self.send_command(enc_command)
+        try:
+            d = await asyncio.wait_for(
+                self.send_privileged_command("reboot"), timeout=timeout
+            )
+        except (asyncio.CancelledError, asyncio.TimeoutError):
+            return {}
+        else:
+            return d
 
     async def factory_reset(self) -> dict:
         """Reset the miner to factory defaults.
@@ -521,14 +511,10 @@ class BTMinerAPI(BaseMinerAPI):
             <summary>Expand</summary>
 
         Returns:
-
             A reply informing of the status of the reset.
         </details>
         """
-        command = {"cmd": "factory_reset"}
-        token_data = await self.get_token()
-        enc_command = create_privileged_cmd(token_data, command)
-        return await self.send_command(enc_command)
+        return await self.send_privileged_command("factory_reset")
 
     async def update_pwd(self, old_pwd: str, new_pwd: str) -> dict:
         """Update the admin user's password.
@@ -545,7 +531,6 @@ class BTMinerAPI(BaseMinerAPI):
             old_pwd: The old admin password.
             new_pwd: The new password to set.
         Returns:
-
             A reply informing of the status of setting the password.
         """
         self.pwd = old_pwd
@@ -555,16 +540,32 @@ class BTMinerAPI(BaseMinerAPI):
                 f"New password too long, the max length is 8.  "
                 f"Password size: {len(new_pwd.encode('utf-8'))}"
             )
-        command = {"cmd": "update_pwd", "old": old_pwd, "new": new_pwd}
-        token_data = await self.get_token()
-        enc_command = create_privileged_cmd(token_data, command)
         try:
-            data = await self.send_command(enc_command)
+            data = await self.send_privileged_command(
+                "update_pwd", old=old_pwd, new=new_pwd
+            )
         except APIError as e:
             raise e
         self.pwd = new_pwd
         return data
 
+    async def net_config(
+        self,
+        ip: str = None,
+        mask: str = None,
+        gate: str = None,
+        dns: str = None,
+        host: str = None,
+        dhcp: bool = True,
+    ):
+        if dhcp:
+            return await self.send_privileged_command("net_config", param="dhcp")
+        if None in [ip, mask, gate, dns, host]:
+            raise APIError("Incorrect parameters.")
+        return await self.send_privileged_command(
+            "net_config", ip=ip, mask=mask, gate=gate, dns=dns, host=host
+        )
+
     async def set_target_freq(self, percent: int) -> dict:
         """Update the target frequency.
 
@@ -578,7 +579,6 @@ class BTMinerAPI(BaseMinerAPI):
         Parameters:
             percent: The frequency % to set.
         Returns:
-
             A reply informing of the status of setting the frequency.
         </details>
         """
@@ -588,10 +588,9 @@ class BTMinerAPI(BaseMinerAPI):
                 f"range.  Please set a % between -10 and "
                 f"100"
             )
-        command = {"cmd": "set_target_freq", "percent": str(percent)}
-        token_data = await self.get_token()
-        enc_command = create_privileged_cmd(token_data, command)
-        return await self.send_command(enc_command)
+        return await self.send_privileged_command(
+            "set_target_freq", percent=str(percent)
+        )
 
     async def enable_fast_boot(self) -> dict:
         """Turn on fast boot.
@@ -603,14 +602,10 @@ class BTMinerAPI(BaseMinerAPI):
         the miner using the Whatsminer tool.
 
         Returns:
-
             A reply informing of the status of enabling fast boot.
         </details>
         """
-        command = {"cmd": "enable_btminer_fast_boot"}
-        token_data = await self.get_token()
-        enc_command = create_privileged_cmd(token_data, command)
-        return await self.send_command(enc_command)
+        return await self.send_privileged_command("enable_btminer_fast_boot")
 
     async def disable_fast_boot(self) -> dict:
         """Turn off fast boot.
@@ -622,14 +617,10 @@ class BTMinerAPI(BaseMinerAPI):
         the miner using the Whatsminer tool.
 
         Returns:
-
             A reply informing of the status of disabling fast boot.
         </details>
         """
-        command = {"cmd": "disable_btminer_fast_boot"}
-        token_data = await self.get_token()
-        enc_command = create_privileged_cmd(token_data, command)
-        return await self.send_command(enc_command)
+        return await self.send_privileged_command("disable_btminer_fast_boot")
 
     async def enable_web_pools(self) -> dict:
         """Turn on web pool updates.
@@ -641,14 +632,10 @@ class BTMinerAPI(BaseMinerAPI):
         password of the miner using the Whatsminer tool.
 
         Returns:
-
             A reply informing of the status of enabling web pools.
         </details>
         """
-        command = {"cmd": "enable_web_pools"}
-        token_data = await self.get_token()
-        enc_command = create_privileged_cmd(token_data, command)
-        return await self.send_command(enc_command)
+        return await self.send_privileged_command("enable_web_pools")
 
     async def disable_web_pools(self) -> dict:
         """Turn off web pool updates.
@@ -660,14 +647,10 @@ class BTMinerAPI(BaseMinerAPI):
         password of the miner using the Whatsminer tool.
 
         Returns:
-
             A reply informing of the status of disabling web pools.
         </details>
         """
-        command = {"cmd": "disable_web_pools"}
-        token_data = await self.get_token()
-        enc_command = create_privileged_cmd(token_data, command)
-        return await self.send_command(enc_command)
+        return await self.send_privileged_command("disable_web_pools")
 
     async def set_hostname(self, hostname: str) -> dict:
         """Set the hostname of the miner.
@@ -681,17 +664,13 @@ class BTMinerAPI(BaseMinerAPI):
         Parameters:
             hostname: The new hostname to use.
         Returns:
-
             A reply informing of the status of setting the hostname.
         </details>
         """
-        command = {"cmd": "set_hostname", "hostname": hostname}
-        token_data = await self.get_token()
-        enc_command = create_privileged_cmd(token_data, command)
-        return await self.send_command(enc_command)
+        return await self.send_privileged_command("set_hostname", hostname=hostname)
 
     async def set_power_pct(self, percent: int) -> dict:
-        """Set the power percentage of the miner.
+        """Set the power percentage of the miner based on current power.  Used for temporary adjustment.
 
         <details>
             <summary>Expand</summary>
@@ -702,7 +681,6 @@ class BTMinerAPI(BaseMinerAPI):
         Parameters:
             percent: The power percentage to set.
         Returns:
-
             A reply informing of the status of setting the power percentage.
         </details>
         """
@@ -713,10 +691,7 @@ class BTMinerAPI(BaseMinerAPI):
                 f"range.  Please set a % between 0 and "
                 f"100"
             )
-        command = {"cmd": "set_power_pct", "percent": str(percent)}
-        token_data = await self.get_token()
-        enc_command = create_privileged_cmd(token_data, command)
-        return await self.send_command(enc_command)
+        return await self.send_privileged_command("set_power_pct", percent=str(percent))
 
     async def pre_power_on(self, complete: bool, msg: str) -> dict:
         """Configure or check status of pre power on.
@@ -734,7 +709,6 @@ class BTMinerAPI(BaseMinerAPI):
                 * `adjust complete`
                 * `adjust continue`
         Returns:
-
             A reply informing of the status of pre power on.
         </details>
         """
@@ -747,13 +721,133 @@ class BTMinerAPI(BaseMinerAPI):
                 '"adjust continue"]'
             )
         if complete:
-            complete = "true"
-        else:
-            complete = "false"
-        command = {"cmd": "pre_power_on", "complete": complete, "msg": msg}
-        token_data = await self.get_token()
-        enc_command = create_privileged_cmd(token_data, command)
-        return await self.send_command(enc_command)
+            return await self.send_privileged_command(
+                "pre_power_on", complete="true", msg=msg
+            )
+        return await self.send_privileged_command(
+            "pre_power_on", complete="false", msg=msg
+        )
+
+    ### ADDED IN V2.0.5 Whatsminer API ###
+    @api_min_version("2.0.5")
+    async def set_temp_offset(self, temp_offset: int) -> dict:
+        """Set the offset of miner hash board target temperature.
+
+        <details>
+            <summary>Expand</summary>
+
+        Set the offset of miner hash board target temperature, only works after
+        changing the password of the miner using the Whatsminer tool.
+
+        Parameters:
+            temp_offset: Target temperature offset.
+        Returns:
+            A reply informing of the status of setting temp offset.
+        </details>
+
+        """
+        if not -30 < temp_offset < 0:
+            raise APIError(
+                f"Temp offset is outside of the allowed "
+                f"range.  Please set a number between -30 and "
+                f"0."
+            )
+
+        return await self.send_privileged_command(
+            "set_temp_offset", temp_offset=temp_offset
+        )
+
+    @api_min_version("2.0.5")
+    async def adjust_power_limit(self, power_limit: int) -> dict:
+        """Set the upper limit of the miner's power. Cannot be higher than the ordinary power of the machine.
+
+        <details>
+            <summary>Expand</summary>
+
+        Set the upper limit of the miner's power, only works after
+        changing the password of the miner using the Whatsminer tool.
+        The miner will reboot after this is set.
+
+        Parameters:
+            power_limit: New power limit.
+        Returns:
+            A reply informing of the status of setting power limit.
+        </details>
+
+        """
+        return await self.send_privileged_command(
+            "adjust_power_limit", power_limit=str(power_limit)
+        )
+
+    @api_min_version("2.0.5")
+    async def adjust_upfreq_speed(self, upfreq_speed: int) -> dict:
+        """Set the upfreq speed, 0 is the normal speed, 9 is the fastest speed.
+
+        <details>
+            <summary>Expand</summary>
+
+        Set the upfreq speed, 0 is the normal speed, 9 is the fastest speed, only works after
+        changing the password of the miner using the Whatsminer tool.
+        The faster the speed, the greater the final hash rate and power deviation, and the stability
+        may be impacted. Fast boot mode cannot be used at the same time.
+
+        Parameters:
+            upfreq_speed: New upfreq speed.
+        Returns:
+            A reply informing of the status of setting upfreq speed.
+        </details>
+        """
+        if not 0 < upfreq_speed < 9:
+            raise APIError(
+                f"Upfreq speed is outside of the allowed "
+                f"range.  Please set a number between 0 (Normal) and "
+                f"9 (Fastest)."
+            )
+        return await self.send_privileged_command(
+            "adjust_upfreq_speed", upfreq_speed=upfreq_speed
+        )
+
+    @api_min_version("2.0.5")
+    async def set_poweroff_cool(self, poweroff_cool: bool) -> dict:
+        """Set whether to cool the machine when mining is stopped.
+
+        <details>
+            <summary>Expand</summary>
+
+        Set whether to cool the machine when mining is stopped, only works after
+        changing the password of the miner using the Whatsminer tool.
+
+        Parameters:
+            poweroff_cool: Whether to cool the miner during power off mode.
+        Returns:
+            A reply informing of the status of setting power off cooling mode.
+        </details>
+        """
+
+        return await self.send_privileged_command(
+            "set_poweroff_cool", poweroff_cool=int(poweroff_cool)
+        )
+
+    @api_min_version("2.0.5")
+    async def set_fan_zero_speed(self, fan_zero_speed: bool) -> dict:
+        """Sets whether the fan speed supports the lowest 0 speed.
+
+        <details>
+            <summary>Expand</summary>
+
+        Sets whether the fan speed supports the lowest 0 speed, only works after
+        changing the password of the miner using the Whatsminer tool.
+
+        Parameters:
+            fan_zero_speed: Whether the fan is allowed to support 0 speed.
+        Returns:
+            A reply informing of the status of setting fan minimum speed.
+        </details>
+
+        """
+        return await self.send_privileged_command(
+            "set_fan_zero_speed", fan_zero_speed=int(fan_zero_speed)
+        )
 
     #### END privileged COMMANDS ####
 
@@ -763,7 +857,6 @@ class BTMinerAPI(BaseMinerAPI):
             <summary>Expand</summary>
 
         Returns:
-
             Summary status of the miner.
         </details>
         """
@@ -775,7 +868,6 @@ class BTMinerAPI(BaseMinerAPI):
             <summary>Expand</summary>
 
         Returns:
-
             Pool status of the miner.
         </details>
         """
@@ -787,7 +879,6 @@ class BTMinerAPI(BaseMinerAPI):
             <summary>Expand</summary>
 
         Returns:
-
             Data on each PGA/ASC with their details.
         </details>
         """
@@ -799,7 +890,6 @@ class BTMinerAPI(BaseMinerAPI):
             <summary>Expand</summary>
 
         Returns:
-
             Data on each PGA/ASC with their details.
         </details>
         """
@@ -811,7 +901,6 @@ class BTMinerAPI(BaseMinerAPI):
             <summary>Expand</summary>
 
         Returns:
-
             Data on all devices with their static details.
         </details>
         """
@@ -823,7 +912,6 @@ class BTMinerAPI(BaseMinerAPI):
             <summary>Expand</summary>
 
         Returns:
-
             Data on the PSU and power information.
         </details>
         """
@@ -839,7 +927,6 @@ class BTMinerAPI(BaseMinerAPI):
         with the other miner APIs.
 
         Returns:
-
             Version data for the miner.
         </details>
         """
@@ -851,7 +938,6 @@ class BTMinerAPI(BaseMinerAPI):
             <summary>Expand</summary>
 
         Returns:
-
             Version data for the miner.
         </details>
         """
@@ -863,7 +949,6 @@ class BTMinerAPI(BaseMinerAPI):
             <summary>Expand</summary>
 
         Returns:
-
             BTMiner status and firmware version.
         </details>
         """
@@ -875,8 +960,21 @@ class BTMinerAPI(BaseMinerAPI):
             <summary>Expand</summary>
 
         Returns:
-
             General miner info.
         </details>
         """
-        return await self.send_command("get_miner_info")
+        return await self.send_command("get_miner_info", allow_warning=False)
+
+    @api_min_version("2.0.1")
+    async def get_error_code(self) -> dict:
+        """Get a list of error codes from the miner.
+
+        <details>
+            <summary>Expand</summary>
+        Get a list of error codes from the miner.  Replaced `summary` as the location of error codes with API version 2.0.4.
+
+        Returns:
+            A list of error codes on the miner.
+        </details>
+        """
+        return await self.send_command("get_error_code", allow_warning=False)
