@@ -1,16 +1,18 @@
-#  Copyright 2022 Upstream Data Inc
-#
-#  Licensed under the Apache License, Version 2.0 (the "License");
-#  you may not use this file except in compliance with the License.
-#  You may obtain a copy of the License at
-#
-#      http://www.apache.org/licenses/LICENSE-2.0
-#
-#  Unless required by applicable law or agreed to in writing, software
-#  distributed under the License is distributed on an "AS IS" BASIS,
-#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
-#  See the License for the specific language governing permissions and
-#  limitations under the License.
+# ------------------------------------------------------------------------------
+#  Copyright 2022 Upstream Data Inc                                            -
+#                                                                              -
+#  Licensed under the Apache License, Version 2.0 (the "License");             -
+#  you may not use this file except in compliance with the License.            -
+#  You may obtain a copy of the License at                                     -
+#                                                                              -
+#      http://www.apache.org/licenses/LICENSE-2.0                              -
+#                                                                              -
+#  Unless required by applicable law or agreed to in writing, software         -
+#  distributed under the License is distributed on an "AS IS" BASIS,           -
+#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.    -
+#  See the License for the specific language governing permissions and         -
+#  limitations under the License.                                              -
+# ------------------------------------------------------------------------------
 
 from pyasic.API import BaseMinerAPI
 
@@ -18,13 +20,14 @@ from pyasic.API import BaseMinerAPI
 class UnknownAPI(BaseMinerAPI):
     """An abstraction of an API for a miner which is unknown.
 
-    This class is designed to try to be a intersection of as many miner APIs
+    This class is designed to try to be an intersection of as many miner APIs
     and API commands as possible (API ⋂ API), to ensure that it can be used
     with as many APIs as possible.
     """
 
-    def __init__(self, ip, port=4028):
+    def __init__(self, ip, api_ver: str = "0.0.0", port: int = 4028):
         super().__init__(ip, port)
+        self.api_ver = api_ver
 
     async def asccount(self) -> dict:
         return await self.send_command("asccount")
