@@ -1,26 +1,35 @@
-#  Copyright 2022 Upstream Data Inc
-#
-#  Licensed under the Apache License, Version 2.0 (the "License");
-#  you may not use this file except in compliance with the License.
-#  You may obtain a copy of the License at
-#
-#      http://www.apache.org/licenses/LICENSE-2.0
-#
-#  Unless required by applicable law or agreed to in writing, software
-#  distributed under the License is distributed on an "AS IS" BASIS,
-#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
-#  See the License for the specific language governing permissions and
-#  limitations under the License.
+# ------------------------------------------------------------------------------
+#  Copyright 2022 Upstream Data Inc                                            -
+#                                                                              -
+#  Licensed under the Apache License, Version 2.0 (the "License");             -
+#  you may not use this file except in compliance with the License.            -
+#  You may obtain a copy of the License at                                     -
+#                                                                              -
+#      http://www.apache.org/licenses/LICENSE-2.0                              -
+#                                                                              -
+#  Unless required by applicable law or agreed to in writing, software         -
+#  distributed under the License is distributed on an "AS IS" BASIS,           -
+#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.    -
+#  See the License for the specific language governing permissions and         -
+#  limitations under the License.                                              -
+# ------------------------------------------------------------------------------
 
-from dataclasses import dataclass, asdict
-from typing import Literal, List
+import logging
 import random
 import string
+import time
+from dataclasses import asdict, dataclass, fields
+from enum import IntEnum
+from typing import List, Literal
 
 import toml
 import yaml
-import json
-import time
+
+
+class X19PowerMode(IntEnum):
+    Normal = 0
+    Sleep = 1
+    LPM = 3
 
 
 @dataclass
@@ -37,6 +46,10 @@ class _Pool:
     username: str = ""
     password: str = ""
 
+    @classmethod
+    def fields(cls):
+        return fields(cls)
+
     def from_dict(self, data: dict):
         """Convert raw pool data as a dict to usable data and save it to this class.
 
@@ -78,6 +91,32 @@ class _Pool:
         pool = {"url": self.url, "user": username, "pass": self.password}
         return pool
 
+    def as_x17(self, user_suffix: str = None) -> dict:
+        """Convert the data in this class to a dict usable by an X5 device.
+
+        Parameters:
+             user_suffix: The suffix to append to username.
+        """
+        username = self.username
+        if user_suffix:
+            username = f"{username}{user_suffix}"
+
+        pool = {"url": self.url, "user": username, "pass": self.password}
+        return pool
+
+    def as_goldshell(self, user_suffix: str = None) -> dict:
+        """Convert the data in this class to a dict usable by a goldshell device.
+
+        Parameters:
+             user_suffix: The suffix to append to username.
+        """
+        username = self.username
+        if user_suffix:
+            username = f"{username}{user_suffix}"
+
+        pool = {"url": self.url, "user": username, "pass": self.password}
+        return pool
+
     def as_inno(self, user_suffix: str = None) -> dict:
         """Convert the data in this class to a dict usable by an Innosilicon device.
 
@@ -136,6 +175,10 @@ class _PoolGroup:
     group_name: str = None
     pools: List[_Pool] = None
 
+    @classmethod
+    def fields(cls):
+        return fields(cls)
+
     def __post_init__(self):
         if not self.group_name:
             self.group_name = "".join(
@@ -171,30 +214,80 @@ class _PoolGroup:
             pools.append(pool.as_x19(user_suffix=user_suffix))
         return pools
 
+    def as_x17(self, user_suffix: str = None) -> dict:
+        """Convert the data in this class to a list usable by an X17 device.
+
+        Parameters:
+             user_suffix: The suffix to append to username.
+        """
+        pools = {
+            "_ant_pool1url": "",
+            "_ant_pool1user": "",
+            "_ant_pool1pw": "",
+            "_ant_pool2url": "",
+            "_ant_pool2user": "",
+            "_ant_pool2pw": "",
+            "_ant_pool3url": "",
+            "_ant_pool3user": "",
+            "_ant_pool3pw": "",
+        }
+        for idx, pool in enumerate(self.pools[:3]):
+            pools[f"_ant_pool{idx+1}url"] = pool.as_x17(user_suffix=user_suffix)["url"]
+            pools[f"_ant_pool{idx+1}user"] = pool.as_x17(user_suffix=user_suffix)[
+                "user"
+            ]
+            pools[f"_ant_pool{idx+1}pw"] = pool.as_x17(user_suffix=user_suffix)["pass"]
+
+        return pools
+
+    def as_goldshell(self, user_suffix: str = None) -> list:
+        """Convert the data in this class to a list usable by a goldshell device.
+
+        Parameters:
+             user_suffix: The suffix to append to username.
+        """
+        return [pool.as_goldshell(user_suffix=user_suffix) for pool in self.pools[:3]]
+
     def as_inno(self, user_suffix: str = None) -> dict:
         """Convert the data in this class to a list usable by an Innosilicon device.
 
         Parameters:
              user_suffix: The suffix to append to username.
         """
-        pools = {}
+        pools = {
+            "Pool1": None,
+            "UserName1": None,
+            "Password1": None,
+            "Pool2": None,
+            "UserName2": None,
+            "Password2": None,
+            "Pool3": None,
+            "UserName3": None,
+            "Password3": None,
+        }
         for idx, pool in enumerate(self.pools[:3]):
             pool_data = pool.as_inno(user_suffix=user_suffix)
             for key in pool_data:
                 pools[f"{key}{idx+1}"] = pool_data[key]
         return pools
 
-    def as_wm(self, user_suffix: str = None) -> List[dict]:
-        """Convert the data in this class to a list usable by an Whatsminer device.
+    def as_wm(self, user_suffix: str = None) -> dict:
+        """Convert the data in this class to a list usable by a Whatsminer device.
 
         Parameters:
              user_suffix: The suffix to append to username.
         """
-        pools = []
-        for pool in self.pools[:3]:
-            pools.append(pool.as_wm(user_suffix=user_suffix))
-        while len(pools) < 3:
-            pools.append({"url": None, "user": None, "pass": None})
+        pools = {}
+        for i in range(1, 4):
+            if i <= len(self.pools):
+                pool_wm = self.pools[i - 1].as_wm(user_suffix)
+                pools[f"pool_{i}"] = pool_wm["url"]
+                pools[f"worker_{i}"] = pool_wm["user"]
+                pools[f"passwd_{i}"] = pool_wm["pass"]
+            else:
+                pools[f"pool_{i}"] = ""
+                pools[f"worker_{i}"] = ""
+                pools[f"passwd_{i}"] = ""
         return pools
 
     def as_avalon(self, user_suffix: str = None) -> str:
@@ -234,7 +327,9 @@ class MinerConfig:
         fan_speed: Manual fan speed to run the fan at (only if temp_mode == "manual").
         asicboost: Whether or not to enable asicboost.
         autotuning_enabled: Whether or not to enable autotuning.
+        autotuning_mode: Autotuning mode, either "wattage" or "hashrate".
         autotuning_wattage: The wattage to use when autotuning.
+        autotuning_hashrate: The hashrate to use when autotuning.
         dps_enabled: Whether or not to enable dynamic power scaling.
         dps_power_step: The amount of power to reduce autotuning by when the miner reaches dangerous temp.
         dps_min_power: The minimum power to reduce autotuning to.
@@ -254,8 +349,11 @@ class MinerConfig:
 
     asicboost: bool = None
 
+    miner_mode: IntEnum = X19PowerMode.Normal
     autotuning_enabled: bool = True
-    autotuning_wattage: int = 900
+    autotuning_mode: Literal["power", "hashrate"] = None
+    autotuning_wattage: int = None
+    autotuning_hashrate: int = None
 
     dps_enabled: bool = None
     dps_power_step: int = None
@@ -263,20 +361,29 @@ class MinerConfig:
     dps_shutdown_enabled: bool = None
     dps_shutdown_duration: float = None
 
+    @classmethod
+    def fields(cls):
+        return fields(cls)
+
     def as_dict(self) -> dict:
         """Convert the data in this class to a dict."""
+        logging.debug(f"MinerConfig - (To Dict) - Dumping Dict config")
         data_dict = asdict(self)
         for key in asdict(self).keys():
+            if isinstance(data_dict[key], IntEnum):
+                data_dict[key] = data_dict[key].value
             if data_dict[key] is None:
                 del data_dict[key]
         return data_dict
 
     def as_toml(self) -> str:
         """Convert the data in this class to toml."""
+        logging.debug(f"MinerConfig - (To TOML) - Dumping TOML config")
         return toml.dumps(self.as_dict())
 
     def as_yaml(self) -> str:
         """Convert the data in this class to yaml."""
+        logging.debug(f"MinerConfig - (To YAML) - Dumping YAML config")
         return yaml.dump(self.as_dict(), sort_keys=False)
 
     def from_raw(self, data: dict):
@@ -286,7 +393,11 @@ class MinerConfig:
         Parameters:
              data: The raw config data to convert.
         """
+        logging.debug(f"MinerConfig - (From Raw) - Loading raw config")
         pool_groups = []
+        if isinstance(data, list):
+            # goldshell config list
+            data = {"pools": data}
         for key in data.keys():
             if key == "pools":
                 pool_groups.append(_PoolGroup().from_dict({"pools": data[key]}))
@@ -301,16 +412,15 @@ class MinerConfig:
                         self.fan_speed = int(data["bitmain-fan-pwm"])
             elif key == "bitmain-work-mode":
                 if data[key]:
-                    if data[key] == 1:
-                        self.autotuning_wattage = 0
+                    self.miner_mode = X19PowerMode(int(data[key]))
             elif key == "fan_control":
-                for _key in data[key].keys():
+                for _key in data[key]:
                     if _key == "min_fans":
                         self.minimum_fans = data[key][_key]
                     elif _key == "speed":
                         self.fan_speed = data[key][_key]
             elif key == "temp_control":
-                for _key in data[key].keys():
+                for _key in data[key]:
                     if _key == "mode":
                         self.temp_mode = data[key][_key]
                     elif _key == "target_temp":
@@ -325,19 +435,25 @@ class MinerConfig:
                     self.asicboost = data[key]["asic_boost"]
 
             if key == "autotuning":
-                for _key in data[key].keys():
+                for _key in data[key]:
                     if _key == "enabled":
                         self.autotuning_enabled = data[key][_key]
                     elif _key == "psu_power_limit":
                         self.autotuning_wattage = data[key][_key]
+                    elif _key == "power_target":
+                        self.autotuning_wattage = data[key][_key]
+                    elif _key == "hashrate_target":
+                        self.autotuning_hashrate = data[key][_key]
+                    elif _key == "mode":
+                        self.autotuning_mode = data[key][_key].replace("_target", "")
 
-            if key == "power_scaling":
-                for _key in data[key].keys():
+            if key in ["power_scaling", "performance_scaling"]:
+                for _key in data[key]:
                     if _key == "enabled":
                         self.dps_enabled = data[key][_key]
                     elif _key == "power_step":
                         self.dps_power_step = data[key][_key]
-                    elif _key == "min_psu_power_limit":
+                    elif _key in ["min_psu_power_limit", "min_power_target"]:
                         self.dps_min_power = data[key][_key]
                     elif _key == "shutdown_enabled":
                         self.dps_shutdown_enabled = data[key][_key]
@@ -348,6 +464,12 @@ class MinerConfig:
         return self
 
     def from_api(self, pools: list):
+        """Convert list output from the `AnyMiner.api.pools()` command into a usable data and save it to this class.
+
+        Parameters:
+            pools: The list of pool data to convert.
+        """
+        logging.debug(f"MinerConfig - (From API) - Loading API config")
         _pools = []
         for pool in pools:
             url = pool.get("URL")
@@ -362,12 +484,19 @@ class MinerConfig:
         Parameters:
             data: The dict config data to convert.
         """
+        logging.debug(f"MinerConfig - (From Dict) - Loading Dict config")
         pool_groups = []
         for group in data["pool_groups"]:
             pool_groups.append(_PoolGroup().from_dict(group))
-        for key in data.keys():
-            if getattr(self, key) and not key == "pool_groups":
+        for key in data:
+            if (
+                hasattr(self, key)
+                and not key == "pool_groups"
+                and not key == "miner_mode"
+            ):
                 setattr(self, key, data[key])
+            if key == "miner_mode":
+                self.miner_mode = X19PowerMode(data[key])
         self.pool_groups = pool_groups
         return self
 
@@ -377,6 +506,7 @@ class MinerConfig:
         Parameters:
             data: The toml config data to convert.
         """
+        logging.debug(f"MinerConfig - (From TOML) - Loading TOML config")
         return self.from_dict(toml.loads(data))
 
     def from_yaml(self, data: str):
@@ -385,15 +515,20 @@ class MinerConfig:
         Parameters:
             data: The yaml config data to convert.
         """
+        logging.debug(f"MinerConfig - (From YAML) - Loading YAML config")
         return self.from_dict(yaml.load(data, Loader=yaml.SafeLoader))
 
-    def as_wm(self, user_suffix: str = None) -> List[dict]:
-        """Convert the data in this class to a config usable by an Whatsminer device.
+    def as_wm(self, user_suffix: str = None) -> dict:
+        """Convert the data in this class to a config usable by a Whatsminer device.
 
         Parameters:
             user_suffix: The suffix to append to username.
         """
-        return self.pool_groups[0].as_wm(user_suffix=user_suffix)
+        logging.debug(f"MinerConfig - (As Whatsminer) - Generating Whatsminer config")
+        return {
+            "pools": self.pool_groups[0].as_wm(user_suffix=user_suffix),
+            "wattage": self.autotuning_wattage,
+        }
 
     def as_inno(self, user_suffix: str = None) -> dict:
         """Convert the data in this class to a config usable by an Innosilicon device.
@@ -401,30 +536,51 @@ class MinerConfig:
         Parameters:
             user_suffix: The suffix to append to username.
         """
+        logging.debug(f"MinerConfig - (As Inno) - Generating Innosilicon config")
         return self.pool_groups[0].as_inno(user_suffix=user_suffix)
 
-    def as_x19(self, user_suffix: str = None) -> str:
+    def as_x19(self, user_suffix: str = None) -> dict:
         """Convert the data in this class to a config usable by an X19 device.
 
         Parameters:
             user_suffix: The suffix to append to username.
         """
+        logging.debug(f"MinerConfig - (As X19) - Generating X19 config")
         cfg = {
-            "pools": self.pool_groups[0].as_x19(user_suffix=user_suffix),
             "bitmain-fan-ctrl": False,
-            "bitmain-fan-pwn": 100,
-            "miner-mode": 0,  # Normal Mode
+            "bitmain-fan-pwn": "100",
+            "freq-level": "100",
+            "miner-mode": str(self.miner_mode.value),
+            "pools": self.pool_groups[0].as_x19(user_suffix=user_suffix),
         }
-        if self.autotuning_wattage == 0:
-            cfg["miner-mode"] = 1  # Sleep Mode
 
         if not self.temp_mode == "auto":
             cfg["bitmain-fan-ctrl"] = True
 
         if self.fan_speed:
-            cfg["bitmain-fan-ctrl"] = str(self.fan_speed)
+            cfg["bitmain-fan-pwn"] = str(self.fan_speed)
 
-        return json.dumps(cfg)
+        return cfg
+
+    def as_x17(self, user_suffix: str = None) -> dict:
+        """Convert the data in this class to a config usable by an X5 device.
+
+        Parameters:
+            user_suffix: The suffix to append to username.
+        """
+        cfg = self.pool_groups[0].as_x17(user_suffix=user_suffix)
+
+        return cfg
+
+    def as_goldshell(self, user_suffix: str = None) -> list:
+        """Convert the data in this class to a config usable by a goldshell device.
+
+        Parameters:
+            user_suffix: The suffix to append to username.
+        """
+        cfg = self.pool_groups[0].as_goldshell(user_suffix=user_suffix)
+
+        return cfg
 
     def as_avalon(self, user_suffix: str = None) -> str:
         """Convert the data in this class to a config usable by an Avalonminer device.
@@ -432,6 +588,7 @@ class MinerConfig:
         Parameters:
             user_suffix: The suffix to append to username.
         """
+        logging.debug(f"MinerConfig - (As Avalon) - Generating AvalonMiner config")
         cfg = self.pool_groups[0].as_avalon(user_suffix=user_suffix)
         return cfg
 
@@ -442,11 +599,12 @@ class MinerConfig:
             model: The model of the miner to be used in the format portion of the config.
             user_suffix: The suffix to append to username.
         """
+        logging.debug(f"MinerConfig - (As BOS) - Generating BOSMiner config")
         cfg = {
             "format": {
                 "version": "1.2+",
-                "model": f"Antminer {model}",
-                "generator": "Upstream Config Utility",
+                "model": f"Antminer {model.replace('j', 'J')}",
+                "generator": "pyasic",
                 "timestamp": int(time.time()),
             },
             "group": [
@@ -463,14 +621,31 @@ class MinerConfig:
         if self.autotuning_enabled or self.autotuning_wattage:
             cfg["autotuning"] = {}
             if self.autotuning_enabled:
-                cfg["autotuning"]["enabled"] = self.autotuning_enabled
-            if self.autotuning_wattage:
-                cfg["autotuning"]["psu_power_limit"] = self.autotuning_wattage
+                cfg["autotuning"]["enabled"] = True
+            else:
+                cfg["autotuning"]["enabled"] = False
+            if self.autotuning_mode:
+                cfg["format"]["version"] = "2.0"
+                cfg["autotuning"]["mode"] = self.autotuning_mode + "_target"
+                if self.autotuning_wattage:
+                    cfg["autotuning"]["power_target"] = self.autotuning_wattage
+                elif self.autotuning_hashrate:
+                    cfg["autotuning"]["hashrate_target"] = self.autotuning_hashrate
+            else:
+                if self.autotuning_wattage:
+                    cfg["autotuning"]["psu_power_limit"] = self.autotuning_wattage
 
         if self.asicboost:
             cfg["hash_chain_global"] = {}
             cfg["hash_chain_global"]["asic_boost"] = self.asicboost
 
+        if self.minimum_fans is not None or self.fan_speed is not None:
+            cfg["fan_control"] = {}
+            if self.minimum_fans is not None:
+                cfg["fan_control"]["min_fans"] = self.minimum_fans
+            if self.fan_speed is not None:
+                cfg["fan_control"]["speed"] = self.fan_speed
+
         if any(
             [
                 getattr(self, item)
@@ -489,7 +664,10 @@ class MinerConfig:
             if self.dps_power_step:
                 cfg["power_scaling"]["power_step"] = self.dps_power_step
             if self.dps_min_power:
-                cfg["power_scaling"]["min_psu_power_limit"] = self.dps_min_power
+                if cfg["format"]["version"] == "2.0":
+                    cfg["power_scaling"]["min_power_target"] = self.dps_min_power
+                else:
+                    cfg["power_scaling"]["min_psu_power_limit"] = self.dps_min_power
             if self.dps_shutdown_enabled:
                 cfg["power_scaling"]["shutdown_enabled"] = self.dps_shutdown_enabled
             if self.dps_shutdown_duration:
